! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module erbackmap_mod 


  use adsmooth_mod
  use param
  use numkind  
  use fitsimage
  use detect_io
  use cal
  use psflib
  use omp_lib

  implicit none

  
  type param_type
    !sequence
    integer(kind=int32) ::  lun
    character(len=1024) ::  image
    character(len=1024) ::  boxlist
    character(len=1024) ::  expimage
    character(len=1024) ::  expimage2 
    character(len=1024) ::  detmask 
    character(len=1024) ::  bkgimage
    character(len=1024) ::  cheesemask 
    integer(kind=int32)     ::  idband
    
    real(kind=real32)       ::  scut
    real(kind=real32)       ::  mlmin
    real(kind=real32)       ::  maxcut   
    integer(kind=int32)     ::  nsplinenodes
    integer(kind=int32)     ::  degree
    logical(kind=4)         ::  smoothflag
    real(kind=real32)       ::  smoothval
    real(kind=real32)       ::  snr
    real(kind=real32)       ::  excesssigma
    integer(kind=int32)     ::  nfitrun
    real(kind=real32)       ::  emin
    real(kind=real32)       ::  emax
    character(len=8)        ::  fitmethod
    logical(kind=4)         ::  expima_flag 
    logical(kind=4)         ::  expima2_flag 
    logical(kind=4)         ::  detmask_flag
    logical(kind=4)         ::  cheesemask_flag     

  end type  param_type
    

contains

!-------------------------------------------------------------------------------

  subroutine erbackmap_ini(par,ierr)

  !---------------------------------------------------------
  ! Calls subroutine OPAST to open parameter and log files
  !---------------------------------------------------------


    type(param_type)    , intent(out) :: par
    integer(kind=int32) , intent(out) :: ierr
    integer(kind=int32)               :: status
    character(len=*),parameter        :: rtname='ERBACKMAP_INI'
    character(len=*),parameter        :: taskname='ERBACKMAP' 

    ierr=0
    status=0

    call opast(taskname,par%lun,status)

    if (status /= 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening parameter file')
    endif 

  end subroutine erbackmap_ini


!-------------------------------------------------------------------------------


  subroutine erbackmap_par(par,ierr)

     
    type(param_type) , intent(out)     :: par
    integer(kind=int32) , intent(out)  :: ierr
    integer(kind=int32)                :: status
    character(len=*),parameter         :: rtname='ERBACKMAP_PAR'

    ierr = 0
    status=0


    call rgetc(par%lun,'IMAGE'     ,par%image       ,status)
    if (status /= 0) ierr = max(ierr,status)
    call rgetc(par%lun,'BOXLIST'   ,par%boxlist     ,status)
    if (status /= 0) ierr = max(ierr,status)      
    call rgetc(par%lun,'BKGIMAGE'   ,par%bkgimage   ,status)
    if (status /= 0) ierr = max(ierr,status)  
    call rgeti(par%lun,'IDBAND'        ,par%idband        ,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetf(par%lun,'SCUT'          ,par%scut          ,status)
    if (status /= 0) ierr = max(ierr,status)      
    call rgetf(par%lun,'MLMIN'         ,par%mlmin         ,status) 
    if (status /= 0) ierr = max(ierr,status)      
    call rgetf(par%lun,'MAXCUT'         ,par%maxcut       ,status) 
    if (status /= 0) ierr = max(ierr,status)        
    call rgeti(par%lun,'NSPLINENODES'  ,par%nsplinenodes  ,status) 
    if (status /= 0) ierr = max(ierr,status)   
    call rgeti(par%lun,'DEGREE'  ,par%degree  ,status) 
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'SMOOTHFLAG'     ,par%smoothflag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgetf(par%lun,'SMOOTHVAL'   ,par%smoothval   ,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetf(par%lun,'SNR'   ,par%snr   ,status) 
    if (status /= 0) ierr = max(ierr,status)    
    call rgetf(par%lun,'EXCESSSIGMA'   ,par%excesssigma   ,status) 
    if (status /= 0) ierr = max(ierr,status)    
    call rgeti(par%lun,'NFITRUN'       ,par%nfitrun       ,status)  
    if (status /= 0) ierr = max(ierr,status)      
    call rgetf(par%lun,'EMIN'         ,par%emin         ,status)  
    if (status /= 0) ierr = max(ierr,status)           
    call rgetf(par%lun,'EMAX'         ,par%emax         ,status)  
    if (status /= 0) ierr = max(ierr,status)                
    call rgetc(par%lun,'FITMETHOD'     ,par%fitmethod     ,status)  
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'EXPIMA_FLAG'     ,par%expima_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)    
    call rgetl(par%lun,'EXPIMA2_FLAG'     ,par%expima2_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetl(par%lun,'DETMASK_FLAG'     ,par%detmask_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetl(par%lun,'CHEESEMASK_FLAG'  ,par%cheesemask_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    if(par%expima_flag .eqv. .true.) then
       call rgetc(par%lun,'EXPIMAGE'  ,par%expimage    ,status)
       if (status /= 0) ierr = max(ierr,status)  
    end if
    call rgetl(par%lun,'EXPIMA2_FLAG'     ,par%expima2_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    if(par%expima2_flag .eqv. .true.) then
       call rgetc(par%lun,'EXPIMAGE2'  ,par%expimage2    ,status)
       if (status /= 0) ierr = max(ierr,status)
    end if
    call rgetl(par%lun,'DETMASK_FLAG'     ,par%detmask_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    if(par%detmask_flag .eqv. .true.) then
       call rgetc(par%lun,'DETMASK'  ,par%detmask    ,status)
       if (status /= 0) ierr = max(ierr,status)
    end if
    call rgetl(par%lun,'CHEESEMASK_FLAG'  ,par%cheesemask_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    if(par%cheesemask_flag .eqv. .true.) then
       call rgetc(par%lun,'CHEESEMASK'  ,par%cheesemask    ,status)
       if (status /= 0) ierr = max(ierr,status)
    end if


    if (ierr /= 0)  call wrerr(rtname,ierr,'Error getting parameters')

    
       
  end subroutine erbackmap_par


!-------------------------------------------------------------------------------



  subroutine erbackmap_in(par,image,boxlist,expimage,expimage2,detmask,ierr)

    type(param_type),            intent(in)   :: par
    type(i4image_type), intent(out)  :: image
    type(i1image_type), intent(out)  :: detmask
    type(r4image_type), intent(out)  :: expimage,expimage2
    type(boxlist_type),intent(out)  :: boxlist
!    type(boxlist_type), pointer, intent(out)  :: boxlist
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)              :: status,ext
    character(len=*), parameter      :: rtname='ERBACKMAP_IN'

    status=0
    ierr=0
    ext=0


    call fget_i4image(par%image,image,status)
    ierr=max(ierr,status)

    if (par%expima_flag) then    
       call fget_r4image(par%expimage,expimage,status)
       ierr=max(ierr,status)
       if (.not. match_size(image%shapehdr,expimage%shapehdr)) then
         ierr=3
         call wrerr(rtname,ierr,'Error : exposure image size does not match')
       endif
    endif
 
    if (par%expima2_flag) then
      call fget_r4image(par%expimage2,expimage2,status)
      ierr=max(ierr,status)
      if (.not. match_size(image%shapehdr,expimage2%shapehdr)) then
        ierr=3
        call wrerr(rtname,ierr,'Error : exposure image 2 size does not match')
      endif      
    endif


    if (par%detmask_flag) then
      call fget_i1image(par%detmask,detmask,status)
      ierr=max(ierr,status)
      if (.not. match_size(image%shapehdr,detmask%shapehdr)) then
        ierr=3
        call wrerr(rtname,ierr,'Error: detection mask size does not match')
      endif         
    endif

  

    call fget_boxlist(par%boxlist,boxlist,status)  

    ierr=max(ierr,status)        

  end subroutine erbackmap_in





!-------------------------------------------------------------------------------

  subroutine  erbackmap_blind(par,cal_par,boxlist,image,detmask,cheesemask,ierr)

    !------------------------------------------------------------------------
    ! Calculates the cheesed mask by setting the detection mask to zero,
    ! where the surface brightness of the sources in the boxlist exceed
    ! the value given by parameter SCUT
    !------------------------------------------------------------------------  

    type(param_type)    :: par
    type(cal_par_type)  :: cal_par
    type(i4image_type)  :: image    
    type(i1image_type)  :: detmask
    type(i1image_type)  :: cheesemask
    type(boxlist_type)  :: boxlist
    type(psf2d_type)    :: psf
    integer(kind=int32) :: ierr 
    character(len=*),parameter   :: rtname = 'ERBACKMAP_BLIND'
    

    character(len=clong) :: psf_file
    logical (kind=log32) :: survmode
    integer(kind=int8), dimension(:,:), allocatable  :: savemask    
    integer(kind=int32) :: status,irow,nstart,nleft,i,j,ext,irad,ix,iy
    real(kind=real32)   :: surf_lim,cutfac,energy,offax,pix2arcsec,x,y,dx,dy
    real(kind=real32)   :: cutrad_sec,cutrad_pix,cutrad_red,rad2,dist2
    real(kind=real32),dimension(:), allocatable    :: cutrad_save
    real(kind=real32)   :: theta,phi
    real(kind=real32),parameter   :: theta_max=29.99
     integer(kind=int32) :: nthread
    logical(kind=log32) :: first

    ierr=0
    status=0

    allocate(cutrad_save(size(boxlist%data)))
    

    ! determine observation mode (POINTING/SURVEY)
    if (image%stdhdr%obs_mode == 'SURVEY' .or. image%stdhdr%obs_mode =='survey') then
      survmode = .true.
    else if (image%stdhdr%obs_mode == 'SCAN' .or. image%stdhdr%obs_mode =='scan') then
      survmode = .true.
    else if (image%stdhdr%obs_mode == 'SLEW' .or. image%stdhdr%obs_mode =='slew') then
      survmode = .true.
    else if  (image%stdhdr%obs_mode == 'POINTING' .or. image%stdhdr%obs_mode =='pointing') then
      survmode = .false. 
    else 
      survmode = .false. 
      ierr=3
      call wrerr(rtname,ierr,'Invalid observing mode')
    endif

!    write(*,*) image%stdhdr%obs_mode, ' ',survmode

    ! copy detection mask to cheesed mask:
    
    if (par%detmask_flag) then
      call copy_i1image(detmask,cheesemask,status)
      allocate(savemask(cheesemask%shapehdr%naxis1,cheesemask%shapehdr%naxis2))
    else
      allocate(cheesemask%arr(image%shapehdr%naxis1,image%shapehdr%naxis2))
      cheesemask%shapehdr=image%shapehdr
      cheesemask%shapehdr%bitpix=8
      cheesemask%stdhdr=image%stdhdr
      cheesemask%refhdr=image%refhdr
      cheesemask%wcshdr=image%wcshdr
      allocate(savemask(image%shapehdr%naxis1,image%shapehdr%naxis2))
      cheesemask%arr=1
    endif

    savemask = cheesemask%arr

    ! mean energy in PI channels:
    energy=0.5*(par%emin+par%emax)
    ! image pixel scale:
    pix2arcsec = 0.5*(abs(cheesemask%wcshdr%cdelt1)+abs(cheesemask%wcshdr%cdelt2))*3600.


    ! set CAL parameters for PSF:
    call cal_setpar(image%stdhdr,cal_par,status)   

    if (survmode) then
      phi=0.0
      ! get survey PSF:
      cal_par%codenam='2D_PSF'
      cal_par%expression=''
      call cal_r4toex('ENERG',energy,cal_par%expression,status)
      call cal_stringtoex('OBSMODE','SURVEY',cal_par%expression,status)
      

      call cal_filename(cal_par,psf_file,ext,status)

      call get_2dpsf(psf_file,ext,phi,psf,status)
    endif

   
    ! count valid pixels in detector mask:
    nstart = sum(int(cheesemask%arr,kind=int32))

    ! reduce cutout radii, if valid mask area falls below threshold:   
    first=.true. 
    cutfac=1.0
    do while (float(nleft)/float(nstart) < par%maxcut) 
      cheesemask%arr = savemask
      ! loop over source list:

     !! !$OMP PARALLEL DEFAULT(PRIVATE) SHARED(boxlist,cheesemask) 
     !!  nthread=omp_get_num_threads()
     !! !$OMP DO
      do irow=1,boxlist%shapehdr%naxis2
     !   write(*,*) 'irow: ',irow
        if (boxlist%data(irow)%id_band == par%idband .and. boxlist%data(irow)%like >= par%mlmin) then 
       !   write(*,*) 'first:',first
          if (first) then
            if (survmode) then
              !avoid divison by zero and negative values (GL 9/11/2017) :
              surf_lim = par%scut / max(1.0,boxlist%data(irow)%scts)
              call psf_surfrad(psf,surf_lim,.false.,cutrad_sec,status)
            else
              call cal_getoffax(boxlist%data(irow)%ra,boxlist%data(irow)%dec,boxlist%stdhdr,theta,phi,status)
              if (theta > theta_max) then
                ierr=2
                call wrerr(rtname,ierr,'off-axis angle outside calibrated range')
                theta=theta_max
              endif
              ! get pointed PSF:
              cal_par%codenam='2D_PSF'
              cal_par%expression=''             
              call cal_r4toex('ENERG',energy,cal_par%expression,status)
              call cal_r4toex('THETA',theta,cal_par%expression,status) 
                
              call cal_stringtoex('OBSMODE','POINTING',cal_par%expression,status)
              call cal_filename(cal_par,psf_file,ext,status)
              call get_2dpsf(psf_file,ext,phi,psf,status)
              ! get cut radius:
              !avoid divison by zero and negative values (GL 9/11/2017) :
              surf_lim = par%scut / max(1.0,boxlist%data(irow)%scts)
              call psf_surfrad(psf,surf_lim,.false.,cutrad_sec,status)
            endif 

!            cutrad_pix = cutrad_sec / pix2arcsec
            ! multiply by box size:
            cutrad_pix = cutrad_sec * float(boxlist%data(irow)%box_size) / pix2arcsec
            cutrad_save(irow) = cutrad_pix                 
          endif !first
          cutrad_pix = cutrad_save(irow) * cutfac
          ! blank cheesed mask:
          rad2=cutrad_pix*cutrad_pix
          irad=ceiling(cutrad_pix)
          ix=nint(boxlist%data(irow)%x_ima)
          iy=nint(boxlist%data(irow)%y_ima)
      !   !$OMP PARALLEL DEFAULT(PRIVATE) SHARED(boxlist,cheesemask)  
      !   !$OMP DO          
          do j=iy-irad,iy+irad
            if (j >= 1 .and. j <= cheesemask%shapehdr%naxis2) then
              y=real(j)
              dy=y-boxlist%data(irow)%y_ima
              do i=ix-irad,ix+irad
                if (j >= 1 .and. j <= cheesemask%shapehdr%naxis2) then
                  x=real(i)
                  dx=x-boxlist%data(irow)%x_ima
                  if ((dx*dx+dy*dy) <  rad2) then
                    if (i >= 1 .and. i <= cheesemask%shapehdr%naxis1) cheesemask%arr(i,j) = 0
                  endif 
                endif
              enddo ! i 
            endif           
          enddo ! j 
      !   !$OMP END DO
      !   !$OMP END PARALLEL          
        endif
      enddo !irow
    !  !$OMP END DO
    !  !$OMP END PARALLEL
     
      nleft =  sum(int(cheesemask%arr,kind=int32))
      cutfac=cutfac*0.95
      first=.false.   
    enddo ! nleft/nstart
  
    
    deallocate(savemask)
    deallocate(cutrad_save)


  end subroutine erbackmap_blind

!--------------------------------------------------------------------------------

  subroutine erbackmap_bin(par,image,cheesemask,grid,ngrid,ierr)

  !-----------------------------------------------------------------------------
  ! Rebin cheesed input image into (par%nsplinenodes)^2 bins.
  ! For each bin calculate centre of mass and valid area from cheesed
  ! mask. 
  ! Array GRID contains Xcent[pixel], Ycent[pixel], NCOUNTS, AREA[pixel] 
  ! for each bin.   
  ! NGRID is the number of valid (AREA > 0) GRID bins. 
  !-----------------------------------------------------------------------------

    type(param_type)    :: par   
    type(i4image_type)  :: image    
    type(i1image_type)  :: cheesemask
    real(kind=real32),dimension(:,:) :: grid
    real(kind=real32),dimension(:,:), allocatable :: grid0  
    integer(kind=int32) :: ngrid,ierr 

    character(len=*), parameter :: rtname='ERBACKMAP_BIN'

    ! sub-bins: 
    integer(kind=int32) :: cmax, isub, ntot
    real(kind=real32)   :: xcen1,xcen2,ycen1,ycen2,xcen,ycen
    real(kind=real32),dimension(:),allocatable   :: xgrid,ygrid 
    real(kind=real32)   ::  qxbsz,  qybsz , hxbsz,  hybsz
    character(len=8)    ::  crstr
    character(len=80)   ::  regfile
    logical             :: splitflag
    !--------
    integer(kind=int32) :: n,xmin,xmax,ymin,ymax,i,j,ibin,jbin
    integer(kind=int32) :: igrid,ibmin,ibmax,jbmin,jbmax
    real(kind=real32)   :: xbsz,ybsz,staterr


    ! test: bins with > cmax counts will be split in 4 sub-bins:
    cmax=3600
   ! splitflag=.true.
    splitflag=.false.

    ! Search cheese mask for min/max coordinates of valid pixels:

    
  

    ierr=0
    n=par%nsplinenodes

    allocate(grid0(4,n*n))

    allocate(xgrid(n*n))
    allocate(ygrid(n*n))

    xmin=cheesemask%shapehdr%naxis1
    ymin=cheesemask%shapehdr%naxis2
    xmax=1
    ymax=1

    do j=1,cheesemask%shapehdr%naxis2
      do i=1,cheesemask%shapehdr%naxis1
        if (cheesemask%arr(i,j) > 0) then
          xmin=min(xmin,i)
          ymin=min(ymin,j)
          xmax=max(xmax,i)
          ymax=max(ymax,j)
        endif
      enddo  
    enddo  



    ! determine binsizes:
    xbsz = real(xmax-xmin+1)/real(n)
    ybsz = real(ymax-ymin+1)/real(n)  
    !quarter binsize

    hxbsz = real(xmax-xmin+1)/real(2*n)
    hybsz = real(ymax-ymin+1)/real(2*n)  

    qxbsz = real(xmax-xmin+1)/real(4*n)
    qybsz = real(ymax-ymin+1)/real(4*n)  
    
    igrid=0
    do i=1,n
      do j=1,n
        igrid=igrid+1
        xgrid(igrid)=real(j-0.5)*xbsz
        ygrid(igrid)=real(i-0.5)*xbsz
       ! write(*,*) 'xgrid,ygrid:',xgrid(igrid),ygrid(igrid)
      end do
    enddo
   
     
    if (xbsz > 0 .and. ybsz > 0) then 

      ! integrate bins on regular grid0:
      ngrid=0
      igrid=1 
      grid0(:,:) = 0.0
      grid(:,:)  = 0.0
      do j=ymin,ymax
        jbin=ceiling((real(j-ymin)+0.5)/ybsz)
        ! corrected ymax -> xmax (GL 03/11/2015)
        do i=xmin,xmax
          ibin=ceiling((real(i-xmin)+0.5)/xbsz)
          igrid=(jbin-1)*n+ibin
          if (cheesemask%arr(i,j) > 0) then
             grid0(1,igrid) = grid0(1,igrid)+real(i)
             grid0(2,igrid) = grid0(2,igrid)+real(j)
             grid0(3,igrid) = grid0(3,igrid)+real(image%arr(i,j))
             grid0(4,igrid) = grid0(4,igrid)+1.0    
          endif
        enddo  
      enddo  

      if (splitflag) then
      ! split bins containing more than <cmax> counts
      ngrid=0
      do igrid=1,n*n  
        if (grid0(4,igrid) > 0.1) then
          if (grid0(3,igrid)  > cmax) then 
         !   write(*,*)  'grid0(3,igrid), cmax: ',grid0(3,igrid), cmax
            xmin=max(1,int(xgrid(igrid)-hxbsz))
            xmax=min(int(xgrid(igrid)+hxbsz),cheesemask%shapehdr%naxis1)
            ymin=max(1,int(ygrid(igrid)-hybsz))
            ymax=min(int(ygrid(igrid)+hybsz),cheesemask%shapehdr%naxis2)
            do j=ymin,ymax
              jbin=min(2,ceiling((real(j-ymin)+0.5)/hybsz))
              do i=xmin,xmax
                ibin=min(2,ceiling((real(i-xmin)+0.5)/hxbsz))
                isub=(jbin-1)*2+ibin
                !write(*,*) 'jbin, ibin, isub, i,j,xmin,ymin: ', jbin, ibin, isub, i,j, xmin,ymin
                if (cheesemask%arr(i,j) > 0) then
                  !write(*,*) 'ngrid,isub: ',ngrid,isub
                  grid(1,ngrid+isub) = grid(1,ngrid+isub)+real(i)
                  grid(2,ngrid+isub) = grid(2,ngrid+isub)+real(j)
                  grid(3,ngrid+isub) = grid(3,ngrid+isub)+real(image%arr(i,j))
                  grid(4,ngrid+isub) = grid(4,ngrid+isub)+1.0    
                endif
              enddo  
            enddo  
            ngrid=ngrid+4
            !split bin
            !xcen=grid(1,igrid) / grid(4,igrid)
          else
            ngrid=ngrid+1
            grid(1:4,ngrid) =  grid0(1:4,igrid) 
          endif
        endif
      end do    
      ntot=ngrid
      else
        ngrid=n*n
        ntot=ngrid
        grid(1:4,1:ngrid) =  grid0(1:4,1:ngrid)
      endif

      ! calculate mean x,y,z:
      ! keep only bins with area > 0
      ngrid=0
      do igrid=1,ntot
        if (grid(4,igrid) > 0.1) then
          ngrid=ngrid+1
          grid(1,ngrid) = grid(1,igrid) / grid(4,igrid)
          grid(2,ngrid) = grid(2,igrid) / grid(4,igrid)
          if (grid(3,ngrid) > 0.5) then 
            staterr = (grid(3,ngrid))**0.5 / grid(4,igrid)
          else
            staterr = 1. / grid(4,igrid)
          endif
          grid(3,ngrid) = grid(3,igrid) / grid(4,igrid)
          grid(4,ngrid) = staterr       
        endif 
      enddo

!      regfile=trim(par%bkgimage)//'.reg'
!      open(1,file=regfile)
!      write(1,*) '# Region file format: DS9 version 4.1'
!      write(1,*) 'global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1'
!      write(1,*) 'image'
!      do igrid=1,ngrid
!        write(crstr,'(F8.4)') grid(3,igrid)
!        write(1,*) 'point(',grid(1,igrid),',',grid(2,igrid),')  # point=cross text={',trim(crstr),'}'
!      enddo
!      close(1)  

    else
      ierr=3
      call wrerr(rtname,ierr,'No valid bins')
    endif  

!    write(*,*) 'done  erbackmap_bin: nodes=', ngrid

  end subroutine  erbackmap_bin



!--------------------------------------------------------------------------------

  subroutine erbackmap_create(par,image,backmap,ierr)

  !----------------------------------------------------------------------- 
  ! Allocate memory for output map and copy relevant header keywords 
  !-----------------------------------------------------------------------

    type(param_type)            :: par
    type(i4image_type)          :: image
    type(r4image_type)          :: backmap     
    integer(kind=int32)         :: ierr,istat

    character(len=*), parameter :: rtname='ERBACKMAP_CREATE'    

    ierr=0

    allocate(backmap%arr(image%shapehdr%naxis1,image%shapehdr%naxis2), STAT=istat)
   
    if (istat == 0) then
      ! copy header from science image
      backmap%shapehdr  = image%shapehdr
      backmap%stdhdr    = image%stdhdr
      backmap%refhdr    = image%refhdr
      backmap%wcshdr    = image%wcshdr
      
      ! set keywords specific for bkg map:
      backmap%shapehdr%bitpix = -32
      ! set here, definition in type declaration didn't work (GL 18/03/2014):
      backmap%stdhdr%hduvers  = '1.0.0  '
      backmap%stdhdr%hduclass = 'OGIP   '
      backmap%stdhdr%hduclas1 = 'IMAGE  '
      backmap%stdhdr%hduclas2 = 'BKG    '              
    else 
      ierr=3
      call wrerr(rtname,ierr,'Cannot allocate memory')
    endif


  end subroutine  erbackmap_create

!--------------------------------------------------------------------------------  


  subroutine erbackmap_expand(par,image,cheesemask,expimage,backmap,ierr)

  !--------------------------------------------------------------------------
  ! Expand all input images by <next> pixels at each side in order to 
  ! avoid border / image wrapping effects with fft.
  ! Copy stripe of pixel data to expansion region.
  ! Call erbackmap_smooth
  ! Copy relevant part of resulting backgroud map to backmap%arr
  !-------------------------------------------------------------------------

 
    type(param_type)            :: par
    type(i4image_type)          :: image
    type(i1image_type)          :: cheesemask
    type(r4image_type)          :: expimage
    type(r4image_type)          :: backmap
    integer(kind=int32)         :: ierr
 

    type(i4image_type)          :: image_ext
    type(i1image_type)          :: cheesemask_ext
    type(r4image_type)          :: expimage_ext
    type(r4image_type)          :: backmap_ext  

    integer(kind=int32)         :: x1,x2,y1,y2

    integer(kind=int32)         :: status,nxnew,nynew
    integer(kind=int32)         :: next
     
    character(len=*), parameter :: rtname='ERBACKMAP_EXPAND'

    ierr=0
    status=0
    next=150

    if  (image%shapehdr%naxis1 < next .or. image%shapehdr%naxis2 < next) then
      ierr=2
      call wrerr(rtname,ierr,'Expansion region > original image')
    endif
        
      

    nxnew=image%shapehdr%naxis1+2*next
    nynew=image%shapehdr%naxis2+2*next


    image_ext%shapehdr%naxis1=nxnew
    image_ext%shapehdr%naxis2=nynew
    allocate(image_ext%arr(nxnew,nynew))
    image_ext%arr = 0
    
    
    expimage_ext%shapehdr%naxis1=nxnew
    expimage_ext%shapehdr%naxis2=nynew
    allocate(expimage_ext%arr(nxnew,nynew))
    expimage_ext%arr = 0.0
 
    cheesemask_ext%shapehdr%naxis1=nxnew
    cheesemask_ext%shapehdr%naxis2=nynew
    allocate(cheesemask_ext%arr(nxnew,nynew))
    cheesemask_ext%arr = 0
 
    backmap_ext%shapehdr%naxis1=nxnew
    backmap_ext%shapehdr%naxis2=nynew
    allocate(backmap_ext%arr(nxnew,nynew))
    backmap_ext%arr = 0.0
 
!! shift stripe to extension:
!    image_ext%arr(next+1:nxnew-next , next+1:nynew-next)  = image%arr
!    image_ext%arr(1:next , 1:nynew) = image_ext%arr(next+1:2*next , 1:nynew)
!    image_ext%arr(nxnew-next+1:nxnew , 1:nynew) = image_ext%arr(nxnew-2*next+1:nxnew-next , 1:nynew)
!    image_ext%arr(1:nxnew , 1:next) = image_ext%arr(1:nxnew , next+1:2*next)
!    image_ext%arr(1:nxnew , nynew-next+1:nynew) = image_ext%arr(1:nxnew , nynew-2*next+1:nynew-next)

! mirror <next> pixels wide stripe of image about its original border:

    image_ext%arr(next+1:nxnew-next , next+1:nynew-next)  = image%arr
    image_ext%arr(next:1:-1 , 1:nynew) = image_ext%arr(next+1:2*next , 1:nynew)
    image_ext%arr(nxnew:nxnew-next+1:-1 , 1:nynew) = image_ext%arr(nxnew-2*next+1:nxnew-next , 1:nynew)
    image_ext%arr(1:nxnew , next:1:-1) = image_ext%arr(1:nxnew , next+1:2*next)
    image_ext%arr(1:nxnew , nynew:nynew-next+1:-1) = image_ext%arr(1:nxnew , nynew-2*next+1:nynew-next)



! mirror:
    if (par%expima_flag) then
      expimage_ext%arr(next+1:nxnew-next , next+1:nynew-next)  = expimage%arr
      expimage_ext%arr(next:1:-1 , 1:nynew) = expimage_ext%arr(next+1:2*next , 1:nynew)
      expimage_ext%arr(nxnew:nxnew-next+1:-1 , 1:nynew) = expimage_ext%arr(nxnew-2*next+1:nxnew-next , 1:nynew)
      expimage_ext%arr(1:nxnew , next:1:-1) = expimage_ext%arr(1:nxnew , next+1:2*next)
      expimage_ext%arr(1:nxnew , nynew:nynew-next+1:-1) = expimage_ext%arr(1:nxnew , nynew-2*next+1:nynew-next)
    endif

! mirror:   
    cheesemask_ext%arr(next+1:nxnew-next , next+1:nynew-next)  = cheesemask%arr
    cheesemask_ext%arr(next:1:-1 , 1:nynew) = cheesemask_ext%arr(next+1:2*next , 1:nynew)
    cheesemask_ext%arr(nxnew:nxnew-next+1:-1 , 1:nynew) = cheesemask_ext%arr(nxnew-2*next+1:nxnew-next , 1:nynew)
    cheesemask_ext%arr(1:nxnew , next:1:-1) = cheesemask_ext%arr(1:nxnew , next+1:2*next)
    cheesemask_ext%arr(1:nxnew , nynew:nynew-next+1:-1) = cheesemask_ext%arr(1:nxnew , nynew-2*next+1:nynew-next)
    

    call erbackmap_smooth(par,image_ext,cheesemask_ext,expimage_ext,backmap_ext,next,status)
    

    backmap%arr = backmap_ext%arr(next+1:nxnew-next, next+1:nynew-next)




  end  subroutine erbackmap_expand 



!--------------------------------------------------------------------------------

  subroutine erbackmap_smooth(par,image,cheesemask,expimage,backmap,next,ierr)

    type(param_type)            :: par
    type(i4image_type)          :: image

    type(i1image_type)          :: detmask
    type(i1image_type)          :: cheesemask
    type(r4image_type)          :: expimage
    type(r4image_type)          :: backmap
    integer(kind=int32)         :: next
    integer(kind=int32)         :: ierr

    type(r4image_type)          :: testima
    real(kind=real32)           :: sigma,snrmin,snrdiff,snrup,snrlow
    real(kind=real32),parameter :: snreps=1E-2
    integer(kind=int32)         :: nsig,isig,ix,iy,imin,iup,ilow
    integer(kind=int32),dimension(1) :: iminarr        

    real(kind=real32)           :: t1,t2
    integer(kind=int16),dimension(:,:),allocatable   :: mindex
    type(i4image_type)          :: ima_index 

    real(kind=real32),dimension(:,:,:), allocatable :: bkgarr
    real(kind=real32),dimension(:,:,:), allocatable :: snrarr

    real(kind=real32),dimension(:,:), allocatable :: kernel,image1,image2
    real(kind=real32),dimension(:,:), allocatable :: smoothed,exp_smoothed,snr,counts
    real(kind=real32)                             :: knorm
    real(kind=real32)                             :: pi=3.141592654,twopi=6.283185308 ,sqrt2=1.4142135624 ,eps=1.E-9
    integer(kind=int32)         :: status,ixsize,iysize
    character(len=80)           :: filename


    character(len=*), parameter :: rtname='ERBACKMAP_SMOOTH'


    ierr=0
    status=0

    nsig=8
    ! test 10/06/2020:
    !sigma=15.0
    snrmin=par%snr

    sigma=par%smoothval

    ixsize=image%shapehdr%naxis1
    iysize=image%shapehdr%naxis2

    allocate(snr(ixsize,iysize),stat=status)
    allocate(mindex(ixsize,iysize),stat=status)
    allocate(kernel(ixsize,iysize),stat=status)
    allocate(image2(ixsize,iysize),stat=status)
    allocate(smoothed(ixsize,iysize),stat=status)
    allocate(exp_smoothed(ixsize,iysize),stat=status)


!    allocate(bkgarr(ixsize,iysize,nsig))
!    allocate(snrarr(ixsize,iysize,nsig))
    allocate(bkgarr(nsig,ixsize,iysize))
    allocate(snrarr(nsig,ixsize,iysize))



    if (status == 0) then
 


      if (par%expima_flag) then
        image1=real(image%arr)*real(cheesemask%arr)
        ! 08/07/2020, divide by exposure map:
        where(expimage%arr > 0.0) image1 = image1/expimage%arr
        !image2=real(cheesemask%arr)*expimage%arr
        !test:
         image2=real(cheesemask%arr)
        
      else
        image1=real(image%arr)*real(cheesemask%arr)
        image2=real(cheesemask%arr)

      endif
  
   !   write(*,*) ' start convolve '
      !! !$OMP PARALLEL DO PRIVATE(sigma,kernel,knorm,ierr,image1,image2,smoothed,exp_smoothed)
      do isig=1,nsig    
      !  write(*,*) isig
        !write(*,*) 'sigma=',sigma
        call  kgauss(sigma,ixsize,iysize,kernel,knorm,ierr)
        call  fconvolve(sigma,image1,image2,kernel,smoothed,exp_smoothed,ixsize,iysize,ierr)
        ! 08/07/2020, multiply by exposure map:
        if (par%expima_flag) smoothed=smoothed*expimage%arr
        counts = smoothed / knorm 
        
        snr = 0.0 
        where(counts > 0.0) snr   = counts /  sqrt(counts)
        snrarr(isig,:,:)= snr
        bkgarr(isig,:,:)= smoothed/exp_smoothed
        !call erbackmap_testout(bkgarr(isig,:,:),smoothed,exp_smoothed,snr,mindex,isig,ixsize,iysize,next)
        sigma = sigma * sqrt2
        !write(*,*) 'minmax(snr):',minval(snr),maxval(snr) 
      enddo   
     !! !$OMP END PARALLEL DO

   !  write(*,*) ' done convolve '
     
     ! Find SNR value nearest to snrmin:

     mindex = minloc(abs(snrarr-snrmin),dim=1)
 
     !backmap%arr = smoothed(mindex,:,:) / exp_smoothed(mindex,:,:)

     backmap%arr(:,:)=0.0  
     do iy=1,iysize
       do ix=1,ixsize

!     do iy=1116,1136
!       do ix=812,821


         iminarr=minloc(abs(snrarr(:,ix,iy)-snrmin))
         imin=iminarr(1)
         snrdiff = snrarr(imin,ix,iy)-snrmin
         if (snrdiff > 0.0) then
           if (imin == 1  .or. abs(snrdiff) < snreps) then
             backmap%arr(ix,iy) = bkgarr(imin,ix,iy) 
           else 
             snrup=snrdiff
             snrlow = abs(snrmin-snrarr(imin-1,ix,iy))  
             if  ((snrup+snrlow) == 0.0 )  then
               backmap%arr(ix,iy) =  snrlow*bkgarr(imin,ix,iy)
             else
               backmap%arr(ix,iy) =  (snrlow*bkgarr(imin,ix,iy)+snrup*bkgarr(imin-1,ix,iy))/(snrup+snrlow)
             endif
           endif
         else 
           if (imin == nsig .or. abs(snrdiff) < snreps) then
             backmap%arr(ix,iy) = bkgarr(imin,ix,iy) 
           else 
             snrlow=abs(snrdiff)
             snrup = abs(snrarr(imin+1,ix,iy)-snrmin)  
             if  ((snrup+snrlow) == 0.0 ) then
               backmap%arr(ix,iy) =  bkgarr(imin,ix,iy)
             else
               backmap%arr(ix,iy) =  (snrup*bkgarr(imin,ix,iy)+snrlow*bkgarr(imin+1,ix,iy))/(snrup+snrlow)
             endif
           endif        
         endif    
       enddo
     enddo
      
      ! write index image: 
     !isig=0
     !call erbackmap_testout(bkgarr(1,:,:),smoothed,exp_smoothed,snr,mindex,isig,ixsize,iysize,next)
     !allocate(ima_index%arr(ixsize,iysize))
     !ima_index%shapehdr%naxis1=ixsize
     !ima_index%shapehdr%naxis2=iysize
     !ima_index%arr=mindex
     !filename=trim(par%bkgimage)//'.imap'
     !call  fput_i4image(filename,ima_index,status)

   else
       ierr=3
       call wrerr(rtname,ierr,'Cannot allocate memory')  
   endif




  end subroutine erbackmap_smooth


!--------------------------------------------------------------------------------

  subroutine erbackmap_spline(par,grid,ngrid,backmap,ierr)
  
  !-----------------------------------------------------------------------------
  ! - Prepare parameters for subroutines SURFIT and BISPEV  
  ! - Calculate weight for input grid
  ! - Call SURFIT, BISPEV
  ! Input parameters:
  !   PAR  : task parameters
  !   GRID : Array containing  Xcent[pixel], Ycent[pixel], NCOUNTS, AREA[pixel]
  !          as prepared by ERBACKMAP_BIN   
  !   NGRID: Number of valid values in array GRID
  ! Output parameters:
  !   BACKMAP : Background map
  !   IERR    : SASS error code  
  ! Comments:
  !   In final background map  pixel values  are set to >= backmin,
  !   this should be replaced by a more meaningful value (e.g. a fraction of the map median).  
  !-----------------------------------------------------------------------------

    type(param_type)                   :: par
    real(kind=real32), dimension(:,:)  :: grid  
    type(r4image_type)                 :: backmap     
    integer(kind=int32)                :: ngrid,ierr   
    
    character(len=*), parameter        :: rtname='ERBACKMAP_SPLINE'
    integer(kind=int32)                :: n,nc,i
    integer(kind=int32), dimension(2)  :: arrshape
    real(kind=real32)                  :: backmin=1E-5
    character(len=5)                   :: statstr

!   surfit parameters:
    integer(kind=int32)                            :: iopt,m,kx,ky,status,nxest,nyest
    integer(kind=int32)                            :: nmax,nx,ny,kwrk,lwrk1,lwrk2
    integer(kind=real32), dimension(:),allocatable :: iwrk
    integer(kind=real32), dimension(2)             :: ordarr
    real(kind=real32), dimension(:),allocatable    :: w,wrk1,wrk2,tx,ty,c
    real(kind=real32)                              :: xb,xe,yb,ye,fp ,eps,s
!   bispev parameters:

    real(kind=real32), dimension(:),allocatable     ::  xout,yout,zout,cc,ttx,tty,wrk
    integer(kind=int32)                             ::  mx,my,mxy,lwrk
    

 
    ierr = 0

!   surfit parameters:
    n = par%nsplinenodes
    m = ngrid
    kx =  par%degree
    ky = kx
    if (par%smoothflag) then
      iopt=0
    else
      iopt=-1
    endif
    if (par%smoothval >= 0.0) then
      s=par%smoothval
    else
      s=m*abs(par%smoothval)
    endif

    nx=2*(kx+1)
    ny=2*(ky+1) 
    nxest=2*(kx+1)+2+n
    nyest=2*(ky+1)+2+n
    nmax=max(nxest,nyest)+2
    lwrk1 = 1000000
    lwrk2 = 100000
    lwrk  = 1000000
    kwrk  = 10000

    eps=1E-3
   


    xb=1.0
    xe=backmap%shapehdr%naxis1
    yb=1.0
    ye=backmap%shapehdr%naxis2

    allocate( c((nxest-kx-1)*(nyest-ky-1)) )
    allocate( wrk(lwrk) )
    allocate( wrk1(lwrk1) )
    allocate( wrk2(lwrk2) )
    allocate(tx(nmax))
    allocate(ty(nmax))
    allocate( w(m) )
    allocate(iwrk(kwrk))

!   bispev parameters
    mx = backmap%shapehdr%naxis1
    my = backmap%shapehdr%naxis2
    mxy = mx*my
    allocate( xout(mx) )
    allocate( yout(my) )  
    allocate( zout(mxy))  

    arrshape(1) = mx
    arrshape(2) = my    
 
    status=0 

!test weight=1.
    ! w(1:ngrid) = 1.0
    ! set weight to 1/err
    w(1:ngrid) = 1./grid(4,1:ngrid)
    !test
    ! w(1:ngrid)=w(1:ngrid)/2.
    
    


    call surfit(iopt,m,grid(1,1:ngrid),grid(2,1:ngrid),grid(3,1:ngrid),w(1:ngrid),xb,xe,yb,ye,kx,ky,s,nxest,nyest,  &
                nmax,eps,nx,tx,ny,ty,c,fp,wrk1,lwrk1,wrk2,lwrk2,iwrk,kwrk,status)

 
    if (status > 0) then
!      ierr=3 
        ! change to ierr=2 => write out bkg map 
        ierr=2 
       write(statstr,'(I5)') status      
      call wrerr(rtname,ierr,statstr//': Error in spline fit')
    !else
    endif     
    if (ierr < 3) then
      nc=(nx-kx-1)*(ny-ky-1) 
      allocate(cc(nc))
      cc=c(1:nc)
      allocate(ttx(nx))
      allocate(tty(ny))
      ttx=tx(1:nx)
      tty=ty(1:ny)

      do i=1,mx
       xout(i)=real(i)
      enddo
      do i=1,my
       yout(i)=real(i)
      enddo

      status=0

      call bispev(ttx,nx,tty,ny,cc,kx,ky,xout,mx,yout,my,zout,wrk,lwrk, &
                 iwrk,kwrk,status)

      

    if (status > 0) then
!      ierr=3
       ! change to ierr=2 => write out bkg map 
       ierr=2 
       write(statstr,'(I5)') status      
      call wrerr(rtname,ierr,statstr//': Error in BISPEV')
    endif  

      ! background not allowed to be < minback:
      where (zout < backmin) zout = backmin

      deallocate(cc)
      deallocate(ttx)
      deallocate(tty)
      
      ! need to transpose zout, since it is ordered as zout(my*ix-1)+jy)
      ordarr=(/2,1/)
      backmap%arr = reshape(zout,arrshape,order=ordarr)


      if (status > 0) then
        ierr=3
        call wrerr(rtname,ierr,'Cannot reconstruct spline map')
      endif
    endif 

    deallocate(c)
    deallocate(wrk)
    deallocate(wrk1)
    deallocate(wrk2)
    deallocate(w)
    deallocate(tx)
    deallocate(ty)    
    deallocate(iwrk) 
    deallocate(xout)
    deallocate(yout) 
    deallocate(zout)     

  end subroutine  erbackmap_spline

!--------------------------------------------------------------------------------

  subroutine erbackmap_clip(par,backmap,grid,ngrid,chisq,ierr)

  !------------------------------------------------------------------------------
  ! Remove GRID bins, if the differ by more than (par%excessigma * err)  
  ! from fitted background map
  ! Input:
  !   PAR: structure containing task parameters
  !   BACKMAP: structure containing current background map
  ! Input/Output:
  !   GRID(:) : Binned image data (input to spline fits)
  !   NGRID: Number of valid bins in GRID   
  ! Output: 
  !   CHISQ: Total CHI^2 (GRID - BACKMAP)
  !   IERR:  Error status    
  !------------------------------------------------------------------------------
    
    type(param_type)    :: par
    type(r4image_type)  :: backmap
    real(kind=real32), dimension(:,:)      :: grid 
    integer(kind=int32)                    :: ngrid,ierr  
    real(kind=real32)                      :: chisq

    real(kind=real32),dimension(4,ngrid)   :: newgrid
    integer(kind=int32)                    :: inew  

 
    integer(kind=int32)                    :: minbin, ibin , ix, iy
    character(len=*), parameter            :: rtname='ERBACKMAP_CLIP'


    ierr=0

  
    minbin=(par%degree+1)**2
    chisq=0
    
    inew=0
    do ibin=1,ngrid
      ix=max(nint(grid(1,ibin)),1)
      ix=min(ix,backmap%shapehdr%naxis1)
      iy=max(nint(grid(2,ibin)),1)
      iy=min(iy,backmap%shapehdr%naxis2)
      chisq=chisq+((grid(3,ibin)-backmap%arr(ix,iy)))**2/(grid(4,ibin))**2
      if (abs(grid(3,ibin)-backmap%arr(ix,iy)) <= (grid(4,ibin) * par%excesssigma)) then
        inew=inew+1
        newgrid(:,inew)=grid(:,ibin)
      endif
    enddo

    if (inew >= minbin) then
      grid(:,1:inew)=newgrid(:,1:inew)
      ngrid=inew
    else
      ierr=1
      call wrerr(rtname,ierr,'Cannot remove bins')
    endif
      

  end  subroutine erbackmap_clip

!--------------------------------------------------------------------------------

  subroutine erbackmap_mask(par,backmap,detmask,ierr)

    type(param_type)                 :: par 
    type(r4image_type)               :: backmap    
    type(i1image_type)               :: detmask
    integer(kind=int32), intent(out) :: ierr

    integer(kind=int32)              :: status,i,j
    character(len=*), parameter      :: rtname='ERBACKMAP_MASK'

    ierr=0
    status=0
    
    if (.not. detmask%shapehdr%naxis1 == backmap%shapehdr%naxis1) status=status+1
    if (.not. detmask%shapehdr%naxis2 == backmap%shapehdr%naxis2) status=status+1


    if (status == 0) then
      do j=1,detmask%shapehdr%naxis2
        do i=1,detmask%shapehdr%naxis1
          if (detmask%arr(i,j) == 0) backmap%arr(i,j) = 0.0
        enddo
      enddo
    else
      ierr=3
    endif


  end subroutine erbackmap_mask

!--------------------------------------------------------------------------------

  subroutine erbackmap_out(par,backmap,cheesemask,ierr)

    type(param_type)    :: par
    type(i1image_type)  :: cheesemask
    type(r4image_type)  :: backmap
    integer(kind=int32) :: ierr
    integer(kind=int32) :: status
    character(len=*), parameter      :: rtname='ERBACKMAP_OUT'
    
    ierr=0
    status=0

    call  fput_r4image(par%bkgimage,backmap,status)   

    if (status /= 0) then
      ierr = 3
      call wrerr(rtname,ierr,'Error writing background map')
      status=0
    endif
    
    if (par%cheesemask_flag) then
      call  fput_i1image(par%cheesemask,cheesemask,status)
        if (status /= 0) then
          ierr = 3
          call wrerr(rtname,ierr,'Error writing cheese mask')
        endif
     endif
    
  end subroutine  erbackmap_out

!--------------------------------------------------------------------------------


  subroutine erbackmap_cls(par,ierr,status)

    type(param_type) ,   intent(in)    :: par
    integer(kind=int32), intent(out)   :: status
    integer(kind=int32)                :: ierr   
    character(len=*),parameter         :: rtname='ERBACKMAP_CLS'
    character(len=*),parameter         :: taskname='ERBACKMAP'   

    status=0  

    call cpast(taskname,par%lun,ierr,status)

    if(status /= 0) then
      call wrerr(rtname,status,'Error closing parameter/log files')
    endif    

    
  end subroutine  erbackmap_cls

!--------------------------------------------------------------------------------   
  

  subroutine erbackmap_testout(bkg,smoothed,exp_smoothed,snr,mindex,isig,ixsize,iysize,next)

    type(i4image_type)                            :: image
    type(r4image_type)                            :: backmap
    type(r4image_type)                            :: outfile
    type(i4image_type)                            :: ima_index 
    real(kind=real32),dimension(:,:)              :: bkg 
    real(kind=real32),dimension(:,:), allocatable :: smoothed,exp_smoothed,snr
    integer(kind=int16),dimension(:,:),allocatable   :: mindex
    integer(kind=int32)                           :: ixsize,iysize,isig,next
    integer(kind=int32)                           :: status
    character(len=2)            :: istr(8)    
    character(len=80)           :: filename


    status=0
    istr = (/'01','02','03','04','05','06','07','08'/)

    allocate(outfile%arr(ixsize,iysize))
    outfile%shapehdr%naxis1=ixsize
    outfile%shapehdr%naxis2=iysize

    if (isig > 0) then 

      filename='bkg_layer'//istr(isig)//'.fits'
      outfile%arr=bkg
      call fput_r4image(filename,outfile,status)

      filename='smoothed_layer'//istr(isig)//'.fits'
      outfile%arr=smoothed
      call fput_r4image(filename,outfile,status)

      filename='exp_smoothed_layer'//istr(isig)//'.fits'
      outfile%arr=exp_smoothed
      call fput_r4image(filename,outfile,status)

      filename='snr_layer'//istr(isig)//'.fits'
      outfile%arr=snr
      call fput_r4image(filename,outfile,status)

    else
      allocate(ima_index%arr(ixsize-2*next,iysize-2*next))
      ima_index%shapehdr%naxis1=ixsize-2*next
      ima_index%shapehdr%naxis2=iysize-2*next
      filename='sig_index.fits'
      ima_index%arr=mindex(next+1:ixsize-next , next+1:iysize-next)
      ima_index%wcshdr =  image%wcshdr
      call fput_i4image(filename,ima_index,status)
    endif

  end subroutine erbackmap_testout






    
    
  

end module erbackmap_mod
